package bid.ex;

import no.bbs.bankid.netcentric.centralverifier.communication.CAPI;
import org.apache.log4j.Logger;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.stream.Stream;

import static java.lang.String.format;

public class BankIDConfig {

    private static final Logger log = Logger.getLogger(BankIDConfig.class);

    private static final String proxyIP =
            System.getProperty("bankid.callbackBaseIp");

    private static final String callbackBaseUrl =
            System.getProperty("bankid.callbackBaseUrl");

    private static final String bankIDCSAndOriginatorUrl =
            System.getProperty("bankid.cspAndOriginator");
    private static final String bankidTrustStorePath =
            System.getProperty("bankid.trustStorePath");
    private static final String bankidWebAddress =
            System.getProperty("bankid.webAddress");

    private static final String bankidMerchantCertificateFilename =
            System.getProperty("bankid.merchantcertificatefile");
    private static final String bankidMerchantCertificateFilenamePassword =
            System.getProperty("bankid.merchantcertificatefile.password");


    public static String getBankIDCSAndOriginatorUrl() {
        return bankIDCSAndOriginatorUrl;
    }

    public static String getMerchantCallbackBaseUrl() {
        return callbackBaseUrl;
    }

    static CAPI capi;
    static int sessioncontext;


    public static CAPI getCAPI() {
        return capi;
    }
    public static int getBIDSessionContext() {
        return sessioncontext;
    }

    static {

        // Check that the required configuration has been performed:
        if (Stream.of(
                bankidMerchantCertificateFilename,
                bankidWebAddress,
                callbackBaseUrl
        ).anyMatch(s -> s.startsWith("FIXME"))) {
            log.error("Necessary configuration has not been performed! Search the project for `FIXME` to find the " +
                    "necessary configuration points.");
            throw new IllegalStateException("Necessary configuration has not been performed! Search the project for " +
                    "`FIXME` to find the necessary configuration points.");
        }

        try {
            URL url = new URL(callbackBaseUrl);
        } catch (MalformedURLException e) {
            throw new IllegalStateException("Illegal value for callbackBaseUrl", e);
        }

        capi = CapiLib.load();
        String res = capi.Initialize();
        if ((null == res) || !(res.equals("0"))) {
            throw new IllegalStateException(
                    "Unable to initialise BankID C Server. BankID C Server returned errorcode: " + res);
        }
        String[] resarr =
                capi.OpenSession(
                        bankidMerchantCertificateFilename,
                        bankidMerchantCertificateFilenamePassword,
                        null,
                        bankidWebAddress,
                        null,
                        bankidTrustStorePath,
                        "ALL",
                        null,
                        null);

        log.info(format("OpenSession returned status %s context %s", resarr[0], resarr[1]));
        if (!"0".equals(resarr[0])) {
            throw new IllegalStateException("Could not open session in bidcserver " + resarr[0]);
        }

        sessioncontext = Integer.parseInt(resarr[1]);
    }

    private static class CapiLib {
        static CAPI load() {
            try {
                log.info("Loading CAPI library");
                System.loadLibrary("CAPI");
                log.info("Loaded CAPI library");
                return new CAPI();
            } catch (Exception e) {
                log.error("Failed loading CAPI", e);
                throw e;
            } finally {
                log.info("Leaving CAPI loader");
            }
        }
    }
}
