package bid.ex.controller;

import bid.ex.BankIDConfig;
import bid.ex.support.ParallelOrSerialDocumentToSign;
import bid.ex.support.SessionSequenceInfo;
import bid.ex.support.XMLAndXSLDummyData;
import no.bbs.bankid.netcentric.centralverifier.communication.*;
import org.apache.log4j.Logger;
import org.apache.pdfbox.cos.COSArray;
import org.apache.pdfbox.cos.COSDictionary;
import org.apache.pdfbox.cos.COSName;
import org.apache.pdfbox.cos.COSStream;
import org.apache.pdfbox.pdmodel.PDDocument;
import org.apache.pdfbox.pdmodel.PDDocumentCatalog;
import org.apache.pdfbox.pdmodel.interactive.digitalsignature.ExternalSigningSupport;
import org.apache.pdfbox.pdmodel.interactive.digitalsignature.PDSignature;
import org.apache.pdfbox.pdmodel.interactive.digitalsignature.SignatureOptions;
import org.apache.pdfbox.pdmodel.interactive.digitalsignature.visible.PDVisibleSigProperties;
import org.apache.pdfbox.pdmodel.interactive.digitalsignature.visible.PDVisibleSignDesigner;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.multipart.MultipartFile;

import javax.servlet.http.HttpServletRequest;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static bid.ex.BankIDConfig.getBIDSessionContext;
import static bid.ex.BankIDConfig.getCAPI;
import static bid.ex.support.SessionSequenceInfo.NextBankIDStep.WAITING_INIT;
import static bid.ex.support.SessionSequenceInfo.NextBankIDStep.WAITING_TO_BE_STARTED;
import static bid.ex.support.Utils.*;
import static java.util.Optional.ofNullable;

/**
 * Demonstrates signing using BankID Web Client.
 */
@SuppressWarnings("ALL")
@Controller
@RequestMapping(value = "/")
public class BankIDWebClientSigningController {

    private Logger log = Logger.getLogger(BankIDWebClientSigningController.class);

    @RequestMapping(value = "bwc_init_sign", method = RequestMethod.GET)
    public String begin() {
        return "bwc_sign_parameters";
    }

    @RequestMapping(value = "bwc_startup_script", method = RequestMethod.GET)
    public String startUpScript(Model model, @RequestParam("sid") final String sid) {
        final SessionSequenceInfo sessionSequenceInfo = SessionSequenceInfo.getSessionData(sid);
        model.addAttribute("initParameters", sessionSequenceInfo.getBankIDHelperInitParams());
        model.addAttribute("redirAtEndUrls", sessionSequenceInfo.getWebClientRedirAtEndUrls());
        return "bwc_startup_script";
    }

    // Init session
    @RequestMapping(value = "bwc_start_sign", method = RequestMethod.POST)
    public String initBankIDSessionAndStartItAll(
        @RequestParam(value = "ssn", defaultValue = "") String ssn,
        @RequestParam(value = "keyStrokeData", defaultValue = "") String keyStrokeData,
        @RequestParam(value = "frameMode", defaultValue = "iframe") final String frameMode,
        @RequestParam(value = "documentDisplayMode", defaultValue = "interior") final String documentDisplayMode,
        @RequestParam(value = "useCallback", defaultValue = "false") final Boolean callJavaScriptCallbackWhenFinished,
        @RequestParam(value = "signingKind", defaultValue = "Parallel") final String signingKindString,
        @RequestParam(value = "selfAssemblerSeals", defaultValue = "false") final Boolean selfAssemblerSeals,
        @RequestParam(value = "selfAssemblerValidation", defaultValue = "false") final Boolean selfAssemblerValidation,
        @RequestParam(value = "noGraphic", defaultValue = "false") final Boolean noGraphic,
        @RequestParam(value = "englishClient", defaultValue = "false") final Boolean englishClient,
        @RequestParam(value = "useNewCMSAndOCSP", defaultValue = "false") final Boolean useNewCMSAndOCSP,
        @RequestParam(value = "useText", defaultValue = "true") final Boolean useText,
        @RequestParam(value = "multiDoc", defaultValue = "false") final Boolean multiDocument,
        @RequestParam("textToSign") String textToSign,
        @RequestParam("pdfToSign") MultipartFile pdf,
        @RequestParam("pdfToSign1") MultipartFile pdf1,
        @RequestParam("pdfToSign2") MultipartFile pdf2,
        @RequestParam("pdfToSign3") MultipartFile pdf3,
        @RequestParam("pdfToSign4") MultipartFile pdf4,
        @RequestParam("visualMerchantSeal") MultipartFile visualMerchantSeal,
        @RequestParam("visualEndUserSeal") MultipartFile visualEndUserSeal,
        HttpServletRequest request,
        Model model) throws IOException {
        SessionSequenceInfo sessionSequenceInfo = new SessionSequenceInfo(WAITING_TO_BE_STARTED);

        CAPI capi = BankIDConfig.getCAPI();
        int sessionContext = BankIDConfig.getBIDSessionContext();

        final String bankIDClientVersion = "2.1";

        String basePath = getPathUpToLastSlash(request.getRequestURI());
        String callBackForBankIDWebClient = BankIDConfig.getMerchantCallbackBaseUrl() + "/bwc_callback_sign";
        String errorUrl = getURLWithSameHostAndPort(request, basePath + "/bwc_error?sid=" + sessionSequenceInfo.getSid()).toString();
        String okUrl = getURLWithSameHostAndPort(request, basePath + "/bwc_ok?sid=" + sessionSequenceInfo.getSid()).toString();

        capi.RemoveInfoItems(sessionContext);
        capi.SetInfoItem(sessionContext, "action", tb("sign"));
        capi.SetInfoItem(sessionContext, "sid", tb(sessionSequenceInfo.getSid()));
        capi.SetInfoItem(sessionContext, "merchantURL", tb(callBackForBankIDWebClient));
        capi.SetInfoItem(sessionContext, "merchantFEDomain", tb(getDomainFromUrl(request)));
        capi.SetInfoItem(sessionContext, "nexturl", tb("https://nrk.no"));
        capi.SetInfoItem(sessionContext, "clientVersion", tb(bankIDClientVersion));
        capi.SetInfoItem(sessionContext, "showUnderstanding", tb("Y"));
        capi.SetInfoItem(sessionContext, "showConfirmation", tb("Y"));
        capi.SetInfoItem(sessionContext, "docDisplayMode", tb(documentDisplayMode));
        capi.SetInfoItem(sessionContext, "withCredentials", tb("N"));
        capi.SetInfoItem(sessionContext, "cmsFormat", tb("PKCS7_ISO320001"));
        capi.SetInfoItem(sessionContext, "ocspFormat", tb("OCSP_RFC6960_COMPATIBLE"));
        capi.SetInfoItem(sessionContext, "localeId", tb((englishClient != null && englishClient) ? "en" : "nb"));

        capi.SetInfoItem(sessionContext, "useragent", tb(getUserAgent(request)));


        String[] strings = capi.InitSession(sessionContext);
        if (!"0".equals(strings[0])) {
            String[] initSessionErrors = capi.GetRecentError(sessionContext);
            String message = "Error in bidcserver:";
            for (String line : initSessionErrors[1].split("\n")) {
                if (!line.isEmpty()) message += "\n    " + line;
            }
            capi.RemoveInfoItems(sessionContext);
            log.error(message);
            throw new IllegalArgumentException("errcode: " + strings[0] + message);
        }

        capi.RemoveInfoItems(sessionContext);

        String helperUri = strings[1];
        String traceID = strings[2];
        String clientID = strings[3];

        final Map<String, String> bidHelperInitParameters = new HashMap<>();

        if (callJavaScriptCallbackWhenFinished) {
            bidHelperInitParameters.put("callback", "bid2.merchantCallback");
        }
        bidHelperInitParameters.put("containerID", "bankidwebclient");
        bidHelperInitParameters.put("height", "500");
        bidHelperInitParameters.put("width", "600");
        bidHelperInitParameters.put("frameMode", frameMode);
        bidHelperInitParameters.put("cid", clientID);

        if ("redirect".equals(frameMode)) {
            bidHelperInitParameters.put("errorRedirectUrl", errorUrl); // Used by bankidhelper
        }


        final Map<String, String> endUrls = new HashMap<>();
        endUrls.put("errorUrl", errorUrl);
        endUrls.put("okUrl", okUrl);
        int signingKind;
        switch (signingKindString) {
            case "Serial": {
                signingKind = CSerialSigningData.BID_SERIAL;
                break;
            }
            case "SerialEndUserOnly": {
                signingKind = CSerialSigningData.BID_SERIAL_END_USER_ONLY;
                break;
            }
            default: {
                signingKind = CSerialSigningData.BID_PARALLEL;
            }
        }
        sessionSequenceInfo.setSigningKind(signingKind);

        List<MultipartFile> pdfs =
                Stream.of(pdf, pdf1, pdf2, pdf3, pdf4)
                .filter(pdfToSign -> pdfToSign != null && !pdfToSign.isEmpty())
                .collect(Collectors.toList());

        sessionSequenceInfo
            .setTraceId(traceID) // For later call to transactionInfo
            .setBankIDHelperInitParams(toJson(bidHelperInitParameters, "callback", "height", "width"))
            .setOkUrl(okUrl)
            .setWebClientRedirAtEndUrls(toJson(endUrls))
            // for later handling of the "initSign" operation:
            .setDataToSign(buildDocumentToSignList(useText, multiDocument, textToSign, pdfs, signingKind))
            .setNextBankIDStep(WAITING_INIT)
            .setSsn(ssn)
            .setKeyStrokedata(keyStrokeData)
            .setSigningKind(signingKind)
            .setUseNewCMSAndOCSPFormats(useNewCMSAndOCSP) //Serial signing ignores this value
        ;

        model.addAttribute("bidhelperUri", helperUri);
        model.addAttribute("sid", sessionSequenceInfo.getSid());

        if (signingKind != CSerialSigningData.BID_PARALLEL) {
            sessionSequenceInfo
                .setSelfAssemblerSeals(selfAssemblerSeals)
                .setSelfAssemblerValidation(selfAssemblerValidation)
                .setNoGraphicFlag(noGraphic)
                .setVisualMerchantSeal(visualMerchantSeal.getBytes())
                .setVisualEndUserSeal(visualEndUserSeal.getBytes())
            ;
        }
        return "bwc_showbankid";
    }

    @RequestMapping(value = "bwc_ok")
    public String bankIDOkLogin(Model model, @RequestParam("sid") String sid) {
        model.addAttribute("sessionSequenceInfo", SessionSequenceInfo.getSessionData(sid));
        return "bankid_success";
    }

    @RequestMapping(value = "bwc_error")
    public String bankIDFailed(@RequestParam("sid") String sid, Model model) {
        model.addAttribute("sessionSequenceInfo", SessionSequenceInfo.getSessionData(sid));
        return "bankid_error";
    }

    private static SignatureOptions getSignatureOptions(PDDocument originalPdf, InputStream visualSeal, Optional<CVisualSealPosition> sealPosition) throws IOException {
        int page;
        double x;
        double y;
        if (!sealPosition.isPresent()) {
            page = 1;
            x = 0;
            y = 0;
        } else {
            page = sealPosition.get().page;
            x = sealPosition.get().x;
            y = sealPosition.get().y;
        }
        PDVisibleSignDesigner designer =
                new PDVisibleSignDesigner(originalPdf, visualSeal, page);
        designer.coordinates((float) x, (float) y);
        PDVisibleSigProperties properties = new PDVisibleSigProperties();
        properties.setPdVisibleSignature(designer);
        properties.buildSignature();
        SignatureOptions signatureOptions = new SignatureOptions();
        signatureOptions.setVisualSignature(properties);
        return signatureOptions;
    }

    private static PDSignature getPdSignature(String metadata) {

        PDSignature pdSignature = new PDSignature();
        pdSignature.setFilter(PDSignature.FILTER_ADOBE_PPKLITE);
        pdSignature.setSubFilter(PDSignature.SUBFILTER_ETSI_CADES_DETACHED);
        pdSignature.setLocation(getLocalHost());
        pdSignature.setReason(metadata);
        pdSignature.setSignDate(Calendar.getInstance());

        return pdSignature;
    }

    private static String getLocalHost() {
        try {
            return InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException e) {
            return "Unknown";
        }
    }

    // Self-assembler callbacks
    static class CallbackWithSid {

        final String sid;

        public CallbackWithSid(String sid) {
            this.sid = sid;
        }

        CByteRangeWithData buildMerchantSealIncrement(int sessioncontext, int docIndex, CByteRangeWithData pdfData, CVisualSealCertData visualSealCertData) {
            try {
                SessionSequenceInfo sessionSequenceInfo = SessionSequenceInfo.getSessionData(sid);
                CSerialSigningData documentToSign = sessionSequenceInfo.getDataToSign().get(docIndex).serialSigningData;
                try (PDDocument originalPdf = PDDocument.load(pdfData.data);) {
                    SignatureOptions signatureOptions = getSignatureOptions(originalPdf, sessionSequenceInfo.getVisualMerchantSeal(),
                        ofNullable(documentToSign.merchant_seal_position));
                    PDSignature pdSignature = getPdSignature(visualSealCertData.organization);
                    originalPdf.addSignature(pdSignature, signatureOptions);

                    return buildSealIncrement(originalPdf, pdfData.data.length);
                }
            } catch (Exception e) {
                e.printStackTrace();
                throw new RuntimeException("Merchant Failed to create document", e);
            }
        }

        CByteRangeWithData buildEndUserSealIncrement(int sessioncontext, int docIndex, CByteRangeWithData pdfData, CVisualSealCertData visualSealCertData) {
            try {
                SessionSequenceInfo sessionSequenceInfo = SessionSequenceInfo.getSessionData(sid);
                CSerialSigningData documentToSign = sessionSequenceInfo.getDataToSign().get(docIndex).serialSigningData;
                try (PDDocument originalPdf = PDDocument.load(pdfData.data);) {
                    SignatureOptions signatureOptions = getSignatureOptions(originalPdf, sessionSequenceInfo.getVisualEndUserSeal(),
                        ofNullable(documentToSign.end_user_seal_position));
                    PDSignature pdSignature = getPdSignature(visualSealCertData.organization);
                    originalPdf.addSignature(pdSignature, signatureOptions);

                    return buildSealIncrement(originalPdf, pdfData.data.length);
                }
            } catch (Exception e) {
                e.printStackTrace();
                throw new RuntimeException("Merchant failed to create document", e);
            }
        }

        CByteRangeWithData buildSealIncrement(PDDocument originalPdf, int originalDocumentLength) throws IOException {

            ByteArrayOutputStream output = new ByteArrayOutputStream();

            ExternalSigningSupport externalSigning = originalPdf.saveIncrementalForExternalSigning(output);
            byte[] signature = new byte[0];
            externalSigning.setSignature(signature); //Set an empty signature to reserve space for bidserver to sign

            byte[] documentWithSignatureBytes = output.toByteArray();
            try (PDDocument signedPDF = PDDocument.load(documentWithSignatureBytes);) {
                int[] byteRangeArr = signedPDF.getLastSignatureDictionary().getByteRange();

                int endUserSealLength = documentWithSignatureBytes.length - originalDocumentLength;
                byte[] endUserSeal = new byte[endUserSealLength];

                System.arraycopy(documentWithSignatureBytes, originalDocumentLength, endUserSeal, 0, endUserSealLength);

                CByteRangeWithData result = new CByteRangeWithData();
                result.data = endUserSeal;
                result.low0 = originalDocumentLength;
                result.len0 = byteRangeArr[1] - originalDocumentLength;
                result.low1 = byteRangeArr[2];
                result.len1 = byteRangeArr[3];

                return result;
            }
        }

        private CByteRangeWithData buildValidationDataIncrement(int sessioncontext, int docnum, CByteRangeWithData pdfData,
                                                                String endUserFullOcspResponse,
                                                                String[] endUserX509Certs,
                                                                String serverFullOcspResponse,
                                                                String[] merchantX509Certs) {
            try (PDDocument originalPdf = PDDocument.load(pdfData.data)) {

                PDDocumentCatalog catalog = originalPdf.getDocumentCatalog();
                COSDictionary catalogDictionary = catalog.getCOSObject();

                List<byte[]> ocspResponses = new ArrayList<>();
                // This is best practice, all earlier ocsps will be referenced from this newest DSS
                Optional<COSArray> previousOcsps = ofNullable((COSDictionary) catalog.getCOSObject().getDictionaryObject("DSS"))
                    .map(cosDictionary -> (COSArray) cosDictionary.getDictionaryObject("OCSPs"));
                // Best practice is to also add the "Certs", dropped here.
                ocspResponses.add(Base64.getDecoder().decode(endUserFullOcspResponse));
                if (serverFullOcspResponse != null && !serverFullOcspResponse.isEmpty()) {
                    ocspResponses.add(Base64.getDecoder().decode(serverFullOcspResponse));
                }
                COSDictionary dssDictionary = createDssDictionary(ocspResponses, previousOcsps);

                catalogDictionary.setNeedToBeUpdated(true);
                catalogDictionary.setItem(COSName.getPDFName("DSS"), dssDictionary);

                ByteArrayOutputStream output = new ByteArrayOutputStream();

                originalPdf.saveIncremental(output);

                byte[] bytes = output.toByteArray();
                CByteRangeWithData result = new CByteRangeWithData();
                result.data = bytes;
                result.low0 = 0;
                result.len0 = bytes.length;
                result.low1 = bytes.length;
                result.len1 = 0;

                return result;
            } catch (IOException e) {
                e.printStackTrace();
                throw new RuntimeException("Failed to add OCSP Response to document", e);
            } catch (Exception e) {
                e.printStackTrace();
                throw new RuntimeException("Merchant failed to add validation data", e);
            }
        }


        COSDictionary createDssDictionary(Iterable<byte[]> ocspResponses, Optional<COSArray> prevOcsps) throws IOException {
            final COSDictionary dssDictionary = new COSDictionary();
            dssDictionary.setNeedToBeUpdated(true);
            dssDictionary.setName(COSName.TYPE, "DSS");

            if (ocspResponses != null)
                dssDictionary.setItem(COSName.getPDFName("OCSPs"), createArray(ocspResponses, prevOcsps));

            return dssDictionary;
        }

        COSArray createArray(Iterable<byte[]> ocspArr, Optional<COSArray> prevOcsps) throws IOException {
            COSArray array = new COSArray();
            array.setNeedToBeUpdated(true);

            for (byte[] oneOcsp : ocspArr)
                array.add(createStream(oneOcsp));
            prevOcsps.ifPresent(prevs -> array.addAll(prevs));

            return array;
        }

        COSStream createStream(byte[] data) throws IOException {
            COSStream stream = new COSStream();
            stream.setNeedToBeUpdated(true);
            final OutputStream unfilteredStream = stream.createOutputStream();
            unfilteredStream.write(data);
            unfilteredStream.flush();
            unfilteredStream.close();
            return stream;
        }
    }

    @RequestMapping(value = "download", method = RequestMethod.GET)
    public ResponseEntity<byte[]> getDownloadData(
            @RequestParam("sid") String sid,
            @RequestParam("index") int index
    ) throws Exception {
        final SessionSequenceInfo sessionSequenceInfo = SessionSequenceInfo.getSessionData(sid);
        byte[] output = sessionSequenceInfo.getCompletePDFs().get(index);

        HttpHeaders responseHeaders = new HttpHeaders();
        responseHeaders.setContentType(MediaType.valueOf("application/pdf"));
        responseHeaders.set("Content-disposition", String.format("attachment; filename=pades-document-%d.pdf", index));
        responseHeaders.setContentLength(output.length);

        return new ResponseEntity<byte[]>(output, responseHeaders, HttpStatus.OK);
    }

    class BidCServerException extends RuntimeException {
        String errCode;

        public BidCServerException(String[] capiReturn) {
            this.errCode = capiReturn[0];
        }
    }

    private String[] throwIfCapiError(String[] capiReturn) {
        if (!"0".equals(capiReturn[0])) throw new BidCServerException(capiReturn);
        return capiReturn;
    }

    // Handler for callbacks from web client (via COI) and Client Proxy
    @SuppressWarnings("Convert2MethodRef")
    @RequestMapping(value = "bwc_callback_sign", method = RequestMethod.POST)
    @ResponseBody
    public String handleCallbackFromClient(
        @RequestParam("sid") String sid,
        @RequestParam("operation") String operation,
        @RequestParam(value = "encKey", defaultValue = "") String encKey,
        @RequestParam(value = "encAuth", defaultValue = "") String encAuth,
        @RequestParam(value = "encData", defaultValue = "") String encData
    ) throws IOException {
        log.info("sid=" + sid + ", operation=" + operation);
        final SessionSequenceInfo sessionSequenceInfo = SessionSequenceInfo.getSessionData(sid);

        final CAPI capi = getCAPI();
        final int sessionContext = getBIDSessionContext();
        try {
            switch (operation) {
                case "initSign": {
                    setTraceIdAndFormats(sessionSequenceInfo, capi, sessionContext);
                    String[] ret;
                    if (!sessionSequenceInfo.isSelfAssemblerSeals()) {
                        capi.SetInfoItem(sessionContext, "noGraphic", tb(sessionSequenceInfo.noGraphicFlagIsSet() ?
                            "Y" : "N"));
                    }

                    if (sessionSequenceInfo.isSerialSigning()) {
                        capi.SetInfoItem(sessionContext, "doSerialSigning", tb("Y"));
                    }
                    addInitialDocuments(sessionContext, capi, sessionSequenceInfo);

                    sessionSequenceInfo.getSsn().ifPresent(ssn -> {
                        capi.SetInfoItem(sessionContext, "userID", tb(ssn));
                        sessionSequenceInfo.getKeyStrokedata().ifPresent(keyStrokeData ->
                            capi.SetInfoItem(sessionContext, "userIDKeystrokeData", tb(keyStrokeData)));
                    });

                    if (sessionSequenceInfo.isSelfAssemblerSeals()) {
                        ret = capi.InitTransactionSelfAssembler(sessionContext, encKey, encData, encAuth, operation,
                            sid, new CallbackWithSid(sid)::buildMerchantSealIncrement);
                    } else {
                        ret = capi.InitTransaction(sessionContext, encKey, encData, encAuth, operation, sid);
                    }
                    throwIfCapiError(ret);
                    intermediateGetCapiSessionData(sessionSequenceInfo, capi, sessionContext);
                    return extractReturnString(ret, sessionSequenceInfo, "initSign");
                }
                case "initSignSignMerchantSeal": {
                    setTraceIdAndFormats(sessionSequenceInfo, capi, sessionContext);

                    intermediateSetBackCapiSessionData(sessionSequenceInfo, capi, sessionContext);

                    String[] ret = throwIfCapiError(
                            capi.InitSignSignMerchantSeal(sessionContext, encKey, encData, encAuth, operation, sid)
                    );

                    intermediateGetCapiSessionData(sessionSequenceInfo, capi, sessionContext);

                    return extractReturnString(ret, sessionSequenceInfo, "initSignSignMerchantSeal");
                }
                case "initSignBuildEndUserSeal": {

                    setTraceIdAndFormats(sessionSequenceInfo, capi, sessionContext);

                    intermediateSetBackCapiSessionData(sessionSequenceInfo, capi, sessionContext);

                    String[] ret = throwIfCapiError(
                            capi.InitSignBuildEndUserSeal(sessionContext, encKey, encData, encAuth, operation, sid,
                                    new CallbackWithSid(sid)::buildEndUserSealIncrement)
                    );

                    intermediateGetCapiSessionData(sessionSequenceInfo, capi, sessionContext);

                    return extractReturnString(ret, sessionSequenceInfo, "initSignBuildEndUserSeal");
                }
                case "verifySign": {
                    setTraceIdAndFormats(sessionSequenceInfo, capi, sessionContext);

                    intermediateSetBackCapiSessionData(sessionSequenceInfo, capi, sessionContext);
                    String firstRet;
                    if (sessionSequenceInfo.isSelfAssemblingValidation()) {
                        firstRet = capi.VerifyTransactionRequestSelfAssembler(sessionContext, encKey, encData,
                            encAuth, operation, sid, new CallbackWithSid(sid)::buildValidationDataIncrement);
                    } else {
                        firstRet = capi.VerifyTransactionRequest(sessionContext, encKey, encData, encAuth, operation, sid);
                    }
                    capi.SetInfoItem(sessionContext, "nexturl", tb(sessionSequenceInfo.getOkUrl()));

                    if (!"0".equals(firstRet)) {
                        sessionSequenceInfo.setErrorMsg("verifyTransactionRequestSerial");
                        sessionSequenceInfo.setErrorCode(firstRet);
                        return "errCode=" + firstRet;
                    }

                    String[] ret = throwIfCapiError(capi.VerifyTransactionResponse(sessionContext));

                    intermediateGetCapiSessionData(sessionSequenceInfo, capi, sessionContext);

                    if (sessionSequenceInfo.isSelfAssemblingValidation()) {
                        // We are finished here, so grab the complete pdfs
                        sessionSequenceInfo.setCompletePDFs(buildCompletePdf(capi, sessionSequenceInfo, sessionContext));
                    }

                    return extractReturnString(ret, sessionSequenceInfo, "verifySign");

                }
                case "verifySignAddDSS": {

                    setTraceIdAndFormats(sessionSequenceInfo, capi, sessionContext);

                    intermediateSetBackCapiSessionData(sessionSequenceInfo, capi, sessionContext);

                    String[] ret = throwIfCapiError(
                            capi.VerifySignAddDss(sessionContext, encKey, encData, encAuth, operation, sid)
                    );

                    String retval;
                    intermediateGetCapiSessionData(sessionSequenceInfo, capi, sessionContext);

                    // We are finished here, so grab the complete pdfs
                    sessionSequenceInfo.setCompletePDFs(buildCompletePdf(capi, sessionSequenceInfo, sessionContext));

                    return extractReturnString(ret, sessionSequenceInfo, "initSignBuildEndUserSeal");
                }
                default:
                    throw new IllegalStateException("Illegal operation: " + operation);
                case "handleError": {
                    String[] ret;
                    // Push back session data, just to make it possible to test both serial and parallel handling in bidcserver
                    intermediateSetBackCapiSessionData(sessionSequenceInfo, capi, sessionContext);
                    if (! sessionSequenceInfo.isSerialSigning()) {
                        String requestRetVal = capi.VerifyTransactionRequest(sessionContext, encKey, encData, encAuth, operation, sid);
                        if ("0".equals(requestRetVal)) {
                            String[] errCodes = capi.GetInfoItem(sessionContext, "errorCode");
                            sessionSequenceInfo.setUserPressedCancel(errCodes[1].equals("9001"));
                            capi.SetInfoItem(sessionContext, "nexturl", tb(sessionSequenceInfo.getOkUrl()));
                        }
                        ret = capi.VerifyTransactionResponse(sessionContext);
                    } else {
                        String requestRetval = capi.VerifyTransactionRequestSelfAssembler(sessionContext, encKey,
                            encData, encAuth, operation, sid, null);
                        if ("0".equals(requestRetval)) {
                            String[] errCodes = capi.GetInfoItem(sessionContext, "errorCode");
                            sessionSequenceInfo.setUserPressedCancel(errCodes[1].equals("9001"));
                            capi.SetInfoItem(sessionContext, "nexturl", tb(sessionSequenceInfo.getOkUrl()));
                        }
                        ret = capi.VerifyTransactionResponse(sessionContext);
                    }
                    throwIfCapiError(ret);
                    return extractReturnString(ret, sessionSequenceInfo, "handleError");
                }
            }
        } catch (BidCServerException e) {
            String[] initSessionErrors = capi.GetRecentError(sessionContext);
            String message = "Error in bidcserver:";
            for (String line : initSessionErrors[1].split("\n")) {
                if (!line.isEmpty()) message += "\n    " + line;
            }
            capi.RemoveInfoItems(sessionContext);
            log.error(message);
            return "errCode=" + e.errCode;
        } finally {
            capi.RemoveInfoItems(sessionContext);
        }
    }

    private void intermediateSetBackCapiSessionData(SessionSequenceInfo sessionSequenceInfo, CAPI capi, int sessionContext) {
        if (sessionSequenceInfo.isSerialSigning()) {
            sessionSequenceInfo.getDataToSign().forEach(documentToSign -> capi.SetSerialSignedData(sessionContext, documentToSign.serialSigningData));
        } else {
            sessionSequenceInfo.getDataToSign().forEach(documentToSign -> capi.SetSignedData(sessionContext, documentToSign.parallelData));
        }
    }

    private void intermediateGetCapiSessionData(SessionSequenceInfo sessionSequenceInfo, CAPI capi, int sessionContext) {
        List<ParallelOrSerialDocumentToSign> dataToSign = sessionSequenceInfo.getDataToSign();
        if (sessionSequenceInfo.isSerialSigning()) {
            for (int i = 0; i < dataToSign.size(); ++i) {
                dataToSign.get(i).serialSigningData = capi.GetSerialSignedData(sessionContext, i).value;
            }
        } else {
            for (int i = 0; i < dataToSign.size(); ++i) {
                dataToSign.get(i).parallelData = capi.GetSignedData(sessionContext, i)[1];
            }

        }

    }

    private void addInitialDocuments(int sessionContext, CAPI capi, SessionSequenceInfo sessionSequenceInfo) {
        int[] docNo = {1};
        sessionSequenceInfo.getDataToSign().forEach(documentToSign -> {
                    byte[] datadescription = tb("Document - " + (docNo[0]++));
                    String errcode;
                    switch (documentToSign.kind) {
                        case BSERIAL:
                            CSerialSigningData serialDoc = documentToSign.serialSigningData;
                            errcode = capi.AddDocumentPdfSerialSign(sessionContext,
                                    Base64.getEncoder().encodeToString(serialDoc.original_data.data),
                                    datadescription,
                                    sessionSequenceInfo.getSigningKind(),
                                    serialDoc.merchant_seal_position,
                                    serialDoc.end_user_seal_position);
                            break;
                        case BPDF:
                            errcode = capi.AddDocumentPdf(sessionContext, Base64.getEncoder().encodeToString(documentToSign.pdf), datadescription);
                            break;
                        case BTXT:
                            errcode = capi.AddDocumentText(sessionContext, Base64.getEncoder().encodeToString(tb(documentToSign.textToSign)), datadescription);
                            break;
                        case BXML:
                            errcode = capi.AddDocumentXML(sessionContext, tb(documentToSign.xml), tb(documentToSign.xsl), datadescription);
                            break;
                        default:
                            throw new IllegalStateException("Unknown document kind");
                    }
                    throwOnErr(sessionContext, capi, errcode);
                }
        );
    }

    private void throwOnErr(int sessionContext, CAPI capi, String errcode) {
        if (!errcode.equals("0")) {
            String[] strings = capi.GetRecentError(sessionContext);
            throw new RuntimeException("Error adding documents ("+errcode+") " + Arrays.asList(strings));
        }
    }

    private void setTraceIdAndFormats(SessionSequenceInfo sessionSequenceInfo, CAPI capi, int sessionContext) {
        if (sessionSequenceInfo.isSerialSigning() || sessionSequenceInfo.getUseNewCMSAndOCSPFormats()) {
            capi.SetInfoItem(sessionContext, "cmsFormat", tb("PKCS7_ISO320001"));
            capi.SetInfoItem(sessionContext, "ocspFormat", tb("OCSP_RFC6960_COMPATIBLE"));
            capi.SetInfoItem(sessionContext, "traceID", tb(sessionSequenceInfo.getTraceId()));
        } else {
            capi.SetInfoItem(sessionContext, "traceID", tb(sessionSequenceInfo.getTraceId()));
        }
    }

    private String extractReturnString(String[] ret, SessionSequenceInfo sessionSequenceInfo, String oper) {
        String retval;

        if ("0".equals(ret[0])) {
            retval = ret[1];
        } else {
            retval = "errCode=" + ret[0];
            sessionSequenceInfo.setErrorMsg(oper);
            sessionSequenceInfo.setErrorCode(ret[0]);
        }
        return retval;
    }

    private List<byte[]> buildCompletePdf(CAPI capi, SessionSequenceInfo sessionSequenceInfo, int sessionContext) {
        List<byte[]> pdfList = new ArrayList<>();
        List<ParallelOrSerialDocumentToSign> dataToSign = sessionSequenceInfo.getDataToSign();
        for (int i = 0; i < dataToSign.size(); ++i) {
            CErrOrResult<String> stringCErrOrResult = capi.GetSerialSignedPdf(sessionContext, i);
            byte[] pdf = Base64.getDecoder().decode(stringCErrOrResult.value);
            pdfList.add(pdf);
        }
        return pdfList;
    }

    private static List<ParallelOrSerialDocumentToSign> buildDocumentToSignList(final boolean useText, final boolean multiDocument,
                                                                                final String textToSign, final List<MultipartFile> pdfs,
                                                                                final int signingKind) throws IOException {

        final boolean doPdfSigning = pdfs.size() > 0;
        List<ParallelOrSerialDocumentToSign> dataToSignList = new ArrayList<>();
        if (signingKind == CSerialSigningData.BID_PARALLEL) {
            if (!multiDocument) {
                if (doPdfSigning) {
                    for (MultipartFile pdf : pdfs) {
                        dataToSignList.add(parallelPdfDocument(pdf));
                    }
                } else if (!useText) {
                    dataToSignList.add(parallelXmlDocument());
                } else {
                    ParallelOrSerialDocumentToSign d = parallelTextDocument(textToSign);
                    dataToSignList.add(d);
                }
            } else {
                // Create some data consisting of text, xml and optionally the PDF
                if (doPdfSigning) {
                    for (MultipartFile pdf : pdfs) {
                        dataToSignList.add(parallelPdfDocument(pdf));
                    }
                    dataToSignList.add(parallelXmlDocument());
                    dataToSignList.add(parallelTextDocument(textToSign));
                }
            }
        } else {
            for (MultipartFile pdf : pdfs) {
                CVisualSealPosition merchantPos = new CVisualSealPosition();
                merchantPos.page = 1;
                merchantPos.x = 20.1234;
                merchantPos.y = 40.050;
                CVisualSealPosition endUserSealPos = new CVisualSealPosition();
                endUserSealPos.page = 1;
                endUserSealPos.x = 200.001;
                endUserSealPos.y = 40.050;
                dataToSignList.add(serialPdfDocument(pdf, signingKind, merchantPos, endUserSealPos));
            }
        }
        return dataToSignList;
    }

    private static ParallelOrSerialDocumentToSign serialPdfDocument(
            final MultipartFile pdf, int signingKind, CVisualSealPosition merchantPos, CVisualSealPosition endUserSealPos
    ) throws IOException {
        CSerialSigningData d = new CSerialSigningData();
        d.merchant_seal_position = merchantPos;
        d.end_user_seal_position = endUserSealPos;
        d.original_data = new CByteRangeWithData();
        byte[] bytes = pdf.getBytes();
        d.original_data.data = bytes;
        d.original_data.low0 = 0;
        d.original_data.len0 = bytes.length;
        d.original_data.low1 = bytes.length;
        d.original_data.len1 = 0;

        return ParallelOrSerialDocumentToSign.serialDoc(d);
    }

    private static ParallelOrSerialDocumentToSign parallelTextDocument(final String textToSign) {
        ParallelOrSerialDocumentToSign d = ParallelOrSerialDocumentToSign.textDoc(textToSign);
        return d;
    }

    private static ParallelOrSerialDocumentToSign parallelXmlDocument() {
        ParallelOrSerialDocumentToSign d = ParallelOrSerialDocumentToSign.xmlDoc(XMLAndXSLDummyData.XML_TO_SIGN, XMLAndXSLDummyData.XSL_TO_USE);
        return d;
    }

    private static ParallelOrSerialDocumentToSign parallelPdfDocument(final MultipartFile pdf) throws IOException {
        ParallelOrSerialDocumentToSign d = ParallelOrSerialDocumentToSign.pdfDoc(pdf.getBytes());
        return d;
    }

    /**
     * to byte
     */
    private byte[] tb(String s) {
        return s == null || s.isEmpty() ? new byte[0] : s.getBytes(StandardCharsets.ISO_8859_1);
    }
}
