package bid.ex.support;

import java.io.*;
import java.util.Collections;
import java.util.List;

public class TestSerializable {

        public static void verifyIsSerializable(Serializable o) {
            try {
                readFromBytes(writeToBytes(o), o.getClass());
            } catch (ClassNotFoundException | IOException e) {
                throw new RuntimeException(o.getClass().getSimpleName() + " is not serializable due to " + e.getMessage(), e);
            }
        }

        static byte[] writeToBytes(Serializable x) throws IOException {
            try (ByteArrayOutputStream byteSaved = new ByteArrayOutputStream()) {
                ObjectOutputStream saveTo = new ObjectOutputStream(byteSaved);
                saveTo.writeObject(x);
                return byteSaved.toByteArray();
            }
        }

        static <T extends Serializable> T readFromBytes(byte[] bytes, Class<T> c) throws IOException, ClassNotFoundException {
            try (ByteArrayInputStream byteSaved = new ByteArrayInputStream(bytes)) {
                ObjectInputStream readFrom = new ObjectInputStream(byteSaved);
                return c.cast(readFrom.readObject());
            }
        }

}
