
package bid.ex.support;

import org.apache.log4j.Logger;

import javax.servlet.http.HttpServletRequest;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.UnknownHostException;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Optional;

public class Utils {
    private static final Logger log = Logger.getLogger(Utils.class);

    public static String getLocalHostName() {
        try {
            return InetAddress.getLocalHost().getCanonicalHostName().toLowerCase();
        } catch (UnknownHostException e) {
            throw new IllegalStateException("You are running on an unknown host");
        }
    }

    /**
     * @param basePath a nonnull string
     * @return up to but not with the last forward slash in a path, if no slash, returns inparameter.
     */
    public static String getPathUpToLastSlash(String basePath) {
        final int lastSlash = basePath.lastIndexOf("/");
        if (lastSlash != -1) {
            basePath = basePath.substring(0, lastSlash);
        }
        return basePath;
    }

    public static URL getURLWithSameHostAndPort(HttpServletRequest request, String path) throws MalformedURLException {
        if (!path.startsWith("/")) {
            path = "/" + path;
        }
        final String hostName = request.getServerName();
        final int port = request.getServerPort();
        return new URL("https", hostName, port == 443 ? -1 : port, path);
    }

    /**
     * Gets the value of the User-Agent request header.
     *
     * @param request http request to extract user agent from.
     * @return user agent string.
     */
    public static String getUserAgent(HttpServletRequest request) {
        String userAgent = request.getHeader("User-Agent");
        if (userAgent == null) {
            userAgent = request.getHeader("User-agent");
        }
        if (userAgent == null) {
            userAgent = request.getHeader("user-agent");
        }
        return userAgent;
    }


    /**
     * Converts from Map to a JSON object. The unquotedVariables parameter specifies keys that should not be wrapped in quotes in the resulting
     * JSON string. All other values in the map will be quoted in the result.

     * @return JSON string
     */
    public static String toJson(final Map<String, String> data, String... unquotedVariables) {
        StringBuilder b = new StringBuilder("{\n");
        HashSet<String> noApostrophes = new HashSet<>();
        Collections.addAll(noApostrophes, unquotedVariables);
        boolean first = true;
        for (Map.Entry<String, String> entry : data.entrySet()) {
            if (entry.getValue() == null || entry.getValue().length() == 0) {
                continue;
            }
            if (!first) {
                b.append(",");
            }
            b.append(" \"").append(entry.getKey()).append("\": ");
            if (noApostrophes.contains(entry.getKey())) {
                b.append(entry.getValue());
            } else {
                b.append("\"").append(entry.getValue()).append("\"");
            }
            b.append("\n");
            first = false;
        }
        return b.append(" }").toString();
    }

    public static String getDomainFromUrl(final HttpServletRequest request) {
        final String hostName = request.getServerName();
        final int port = request.getServerPort();
        return "https://" + hostName + (port == 443 ? "" : (":" + port));
    }

    public static String nextSid() {
        return java.util.UUID.randomUUID().toString();
    }

}
