package bid.ex.webclient;

import java.io.BufferedReader;
import java.io.IOException;
import java.net.URLDecoder;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletRequestWrapper;
import org.apache.log4j.Logger;

/**
 * IE sends no content-type or content-type equal to 'text/plain' when it should be application/x-www-form-urlencoded.
 */
public class BankIDWebClientIEFilter implements Filter {
    private static final Logger log = Logger.getLogger(BankIDWebClientIEFilter.class);
    private static final String REOUEST_ENCODING = "UTF-8";


    @Override
    public void init(FilterConfig filterConfig) throws ServletException {

    }

    @Override
    public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException {
        HttpServletRequest servletRequest = (HttpServletRequest) request;
        String ua = servletRequest.getHeader("User-Agent");

        if (servletRequest.getMethod().equals("POST") && ua != null && ua.contains("MSIE") && (request.getContentType() == null ||
                request.getContentType().equals("text/plain"))) {
            final Map<String, String[]> parameters = getDecodedParameters((HttpServletRequest) request);
            log.debug("Received CORS request from MSIE without content-type, assuming application/x-www-form-urlencoded, " +  REOUEST_ENCODING);
            final String[] sids = parameters.get("sid");
            log.debug("sid in params =" + (sids == null || sids.length == 0 ? "null" : sids[0]));
            HttpServletRequestWrapper wrapper = new HttpServletRequestWrapper((HttpServletRequest) request) {
                @Override
                public String getParameter(String name) {
                    final String[] strings = parameters.get(name);
                    return strings == null || strings.length == 0 ? null : strings[0];
                }

                @Override
                public Map<String, String[]> getParameterMap() {
                    return parameters;
                }

                @Override
                public Enumeration<String> getParameterNames() {
                    return Collections.enumeration(parameters.keySet());
                }

                @Override
                public String[] getParameterValues(final String name) {
                    return parameters.get(name);
                }
            };
            chain.doFilter(wrapper, response);
        } else {
            chain.doFilter(request, response);
        }
    }

    public Map<String, String[]> getDecodedParameters(HttpServletRequest request) throws IOException {
        Map<String, List<String>> parameters = new HashMap<>();
        for (Map.Entry<String, String[]> stringArrayEntry : request.getParameterMap().entrySet()) {
            parameters.put(stringArrayEntry.getKey(), Arrays.asList(stringArrayEntry.getValue()));
        }

        BufferedReader reader = request.getReader();
        StringBuilder buffer = new StringBuilder();

        String text;
        while ((text = reader.readLine()) != null) {
            buffer.append(text);
        }
        String[] pairs = buffer.toString().split("\\&");

        for (String pair : pairs) {
            String[] fields = pair.split("=");

            String name = fields.length == 0 ? "" : URLDecoder.decode(fields[0], REOUEST_ENCODING);
            String value = fields.length > 1 ? URLDecoder.decode(fields[1], REOUEST_ENCODING) : "";
            List<String> values = parameters.get(name);
            if (values == null) {
                values = new ArrayList<>();
                parameters.put(name, values);
            }
            values.add(value);
        }
        final Map<String, String[]> returnParameters = new HashMap<>();
        for (Map.Entry<String, List<String>> stringArrayListEntry : parameters.entrySet()) {
            final List<String> value = stringArrayListEntry.getValue();
            returnParameters.put(stringArrayListEntry.getKey(), value.toArray(new String[value.size()]));
        }
        return Collections.unmodifiableMap(returnParameters);
    }

    @Override
    public void destroy() {

    }
}
